// Copyright 2023 T-Systems International GmbH
import app from 'app';
import soaSvc from 'soa/kernel/soaService';
import appCtxSvc from 'js/appCtxService';
import _ from 'lodash';
import messagingSvc from 'js/messagingService';
import dateTimeSvc from 'js/dateTimeService';
import adapterSvc from 'js/adapterService';
import AwPromiseService from 'js/awPromiseService';
import eventBus from 'js/eventBus';
import commandPanelService from 'js/commandPanel.service';
import dms from 'soa/dataManagementService';
import viewModelObjectService from 'js/viewModelObjectService';
import uwPropertySvc from 'js/uwPropertyService';
import popupService from 'js/popupService';
import navigationSvc from 'js/navigationService';
import localStrg from 'js/localStorage';
import localeSvc from 'js/localeService';
import modelPropSvc from 'js/modelPropertyService';
import awSearchService from 'js/awSearchService';
import prefSvc from 'soa/preferenceService';
// import awIconSvc from 'js/awIconService';
// import cfgSvc from 'js/configurationService';
// import commandsSvc from 'js/command.service';
// import AwStateService from 'js/awStateService';
// import cdm from 'soa/kernel/clientDataModel';
// import cmm from 'soa/kernel/clientMetaModel';
// import editHandlerService from 'js/editHandlerService';
// import ClipboardService from 'js/clipboardService';
// import tcClipboardService from 'js/tcClipboardService';

/**
 * TCI specific service for TCI command Handlers
 *
 * @module js/tciService
 */

var exports = {};

/**
 * TCI Version Information
 */
const TCI_VERSION = '3.13.0.0.03';
const TCI_RELEASE_DATE = '2025-02-28T14:55:27.457+0100';
const TCI_COPY_YEAR = '2025';

/**
 * TCI Request Ids/Names
 */
const START_JOB = "StartJob";
const ENOVIA_LOGIN = "EnoviaLogin";
const ENOVIA_QUERY = "EnoviaQuery";
const GET_ROLE_INFO_FOR_NAME = "GetRoleInfoForName";
const GET_PROCESS_INFO_FOR_USER = "GetProcessInfoForUser";
const GET_ENOVIA_QUERY_SETTINGS = "GetEnoviaQuerySettings";
const GET_ITEM_REVISION_PROPERTIES = "GetItemRevisionProperties";
const GET_META_UPDATE_INFO_FOR_UID_REQUEST_NAME = "GetMetaUpdateInfoForUid";
const GET_SETTINGS_INFO = "GetSettingsInfo";
const SEND_ENOVIA_OBJECT_TO_TC = "SendEnoviaObjectToTC";

/**
 * TCI XML Elements/ Attributes
 */
const ID = 'id';
const MUI = "MUI";
const NAME = 'name';
const CODE = "code";
const VALUE = "value";
const ERROR = "Error";
const V6_NAME = "v6Name";
const MESSAGE = "Message";
const DISPLAY = "display";
const PDM_NAME = "pdmName";
const READ_ONLY = "readOnly";
const ATTRIBUTE = "Attribute";
const TC_REV_OBID = "TC_REV_OBID";
const DISPLAY_NAME = "displayName";
const ATTRIBUTE_NAME_USER = "USER";
const ENOVIA_CONNECTOR_TYPE = "ConnectorType";
const ENOVIA_SC_NAME_LIST = "QueryTypeNameList";
const ENOVIA_SC_DISPLAY_LIST = "QueryTypeDisplayList";
const ENOVIA_DEFAULT_ENOVIA_ROLE = "DefaultEnoviaRole";
const ENOVIA_QUERY_TYPE_NAME_LIST = "QueryTypeNameList";
const ENOVIA_DEFAULT_ENOVIA_VAULT = "DefaultEnoviaVault";
const ENOVIA_PREFERRED_CREDENTIALS = "PreferredCredentials";
const ENOVIA_QUERY_TYPE_DISPLAY_LIST = "QueryTypeDisplayList";

const LIST_SEPARATOR = "@EOL@";

//Other Constants
const IS_TCI_LOGGED_IN = 'isLoggedInTo3DX';

//Data Store
let expData;
let expQueryInput;

/**
 * Convert string to Hex String.
 * @param {String} data - text to convert to bytes
 * @returns {String} Hex string for the input string 
 */
let toHexString = (data) => {
	let hexOut = '';
	for (let i = 0; i < data.length; i++) {
		let s = data.charCodeAt(i).toString(16);
		while (s.length < 2) {
			s = '0' + s;
		}
		hexOut += s;
	}

	return hexOut;
};

let publishExpDataToLocalStorage = () => {
	localStrg.publish('expData', JSON.stringify(expData));
};

let updateExpDataFromLocalStorage = () => {
	const expDataJson = localStrg.get('expData');
	expData = JSON.parse(expDataJson);
};


let updateTciLogginFromLocalStorage = (eventData) => {
	let isTciLoggedInValueLocStorage = false; //if value not found in local storage, flag set as a false
	let isTciLoggedInValueLocStorageValue = localStrg.get(IS_TCI_LOGGED_IN);
	if (isTciLoggedInValueLocStorageValue) { //checks value got local storage
		isTciLoggedInValueLocStorage = isTciLoggedInValueLocStorageValue === 'true';
	}

	var isTciLoggedInValueCtx = appCtxSvc.getCtx(IS_TCI_LOGGED_IN);
	if (isTciLoggedInValueLocStorage != isTciLoggedInValueCtx) {
		appCtxSvc.registerCtx(IS_TCI_LOGGED_IN, isTciLoggedInValueLocStorage);
		updateExpDataFromLocalStorage();
	}
}

let setTciLoggedOut = (eventData) => {

	localStrg.publish(IS_TCI_LOGGED_IN, false);
	appCtxSvc.registerCtx(IS_TCI_LOGGED_IN, false);

	eventBus.publish(IS_TCI_LOGGED_IN, false, false);
	localStrg.publish(IS_TCI_LOGGED_IN, false);

	publishExpDataToLocalStorage();
};

let deleteExpdataFromLocalStorage = () => {
	clearCachedData('expdata');
};

let setSubscriptionsForTciLogin = () => {
	let isTciLoggedInValueLocStorageValue = localStrg.get(IS_TCI_LOGGED_IN);
	if (isTciLoggedInValueLocStorageValue) {
		//It is here means the value exist in local storage, it may be true or false
		let isTciLoggedInValueLocStorage = false;
		isTciLoggedInValueLocStorage = isTciLoggedInValueLocStorageValue === 'true';
		appCtxSvc.registerCtx(IS_TCI_LOGGED_IN, isTciLoggedInValueLocStorage);
	} else {
		localStrg.publish(IS_TCI_LOGGED_IN, false);
		appCtxSvc.registerCtx(IS_TCI_LOGGED_IN, false);
	}

	// Subscribe to Teamcenter signOut to clear the localStorage Data
	eventBus.subscribe('session.signOut', setTciLoggedOut);
	eventBus.subscribe('session.signOut', deleteExpdataFromLocalStorage);

	// Subscribe to Teamcenter signIn to login when new page is opened using localStorage Data
	// eventBus.subscribe('session.signIn', updateTciLogginFromLocalStorage); //does not work
	eventBus.subscribe(IS_TCI_LOGGED_IN, updateTciLogginFromLocalStorage);
	localStrg.subscribe(IS_TCI_LOGGED_IN, updateTciLogginFromLocalStorage);

	updateExpDataFromLocalStorage();
	localStrg.subscribe('expData', updateExpDataFromLocalStorage);
}
// setSubscriptionsForTciLogin(); //This does not run on Page load or before opening teh menu, so not useful

//This will run when the tolltip extension on TCU Group Menu is activated, i.e. when the move moved over the TCI Group Menu
let executeInitializeDataFromLocalStorage = true;
export let initializeDataFromLocalStorage = () => {
	if (executeInitializeDataFromLocalStorage) {
		setSubscriptionsForTciLogin();
		executeInitializeDataFromLocalStorage = false;
	}
};


let setTciLoggedIn = () => {
	//Set the value on local storage and Context
	//Local storage is available for all pages and also on refresh
	//Context is available for only current pafe till refresh is preformed
	//The conditions to show the login, logout and other menus read the value from local context
	localStrg.publish(IS_TCI_LOGGED_IN, true);
	appCtxSvc.registerCtx(IS_TCI_LOGGED_IN, true);

	//Publish that the data is changed
	eventBus.publish(IS_TCI_LOGGED_IN, true, false);

	publishExpDataToLocalStorage();
};


/**
 * Dummy alert.
 * @param {String} text - text to display
 */
export let alert2 = function (text) {
	alert(text); // eslint-disable-line
};

let getDefaultRevNVarRules = (ctx) => {
	let rules = {
		revRule: {
			dbValue: "",
			dispValue: ""
		},
		varRule: {
			dbValue: "",
			dispValue: ""
		}
	}

	if (ctx?.occmgmtContext?.productContextInfo) {

		rules.revRule.dbValue = ctx.occmgmtContext.productContextInfo.props.awb0CurrentRevRule.dbValues[0];
		rules.revRule.dispValue = ctx.occmgmtContext.productContextInfo.props.awb0CurrentRevRule.uiValues[0];

		rules.varRule.dbValue = ctx.occmgmtContext.productContextInfo.props.awb0VariantRules.dbValues[0];
		rules.varRule.dispValue = ctx.occmgmtContext.productContextInfo.props.awb0VariantRules.uiValues[0];
	} else {
		rules.revRule.dbValue = ctx.userSession.props.awp0RevRule.dbValue;
		rules.revRule.dispValue = ctx.userSession.props.awp0RevRule.uiValue;
	}

	return rules;
};

export const initializeVMPDataForViewPropsIn3DX = (propArray) => {
	var vmpArray = [];
	_.forEach(propArray, function (node) {
		let prop = {
			name: node.name,
			displayName: node.displayName,
			type: 'STRING',
			isEditable: node.isEditable,
			isRequired: node.isRequired,
			dbValue: node.dbValue,
			uiValue: node.uiValue,
			labelPosition: 'PROPERTY_LABEL_AT_SIDE'
		};
		var vmProp = modelPropSvc.createViewModelProperty(prop);
		uwPropertySvc.setRenderingHint(vmProp, 'textbox');
		uwPropertySvc.setDisplayValue(vmProp, [node.dbValue]);
		vmProp.v6Name = node.v6Name;
		vmProp.name = node.name;
		vmpArray.push(vmProp);
	});
	return vmpArray;
};

export const initializeVMPDataForUpdatePropsIn3DX = (propArray, selectedObject) => {
	var vmpObjectArray = [];
	let dummyUid = 0;
	_.forEach(propArray, function (p) {
		++dummyUid;
		let vmpObject = {
			uid: dummyUid,
			type: "DummyPropertyObject",
			props: {}
		};

		for (const [key, value] of Object.entries(p)) {
			// let dspName = ((key == 'pdmDisplayName' || key == 'v6DisplayName')? 'Name': key);
			// dspName = ((key == 'pdmValue' || key == 'v6Value')? 'Value': dspName);

			// if(key == 'v6Name'){
			// 	value = 'Temp_v6Name';
			// }

			let prop = {
				type: 'STRING',
				hasLov: false,
				isArray: false,
				displayValue: value,
				uiValue: value,
				value: value,
				propertyName: key,
				propertyDisplayName: key,
				isEnabled: true
			};
			// var vmProp = viewModelObjectService.constructViewModelProperty(prop, key, null, false);
			// // var vmProp = modelPropSvc.createViewModelProperty(prop, key);
			// uwPropertySvc.setRenderingHint(vmProp, 'textbox');
			// uwPropertySvc.setDisplayValue(vmProp, [value]);
			// vmProp.v6Name = node.v6Name;
			// vmProp.name = node.name;

			vmpObject.props[key] = prop;
		}
		// const vmo = viewModelObjectService.constructViewModelObject(vmpObject);
		// Object.keys(vmo.props).forEach(function(key) {
		// 	let val = vmo.props[key].value;
		// 	// if(key == 'v6Name'){
		// 	// 	val = 'Temp_v6Name';
		// 	// }
		// 	uwPropertySvc.setDisplayValue(vmo.props[key], [val]);
		// });
		vmpObjectArray.push(vmpObject);
	});
	return vmpObjectArray;
};

let closeCurrentPanel = () => {
	//To close the panel
	var eventData = { source: 'toolAndInfoPanel' };
	eventBus.publish('complete', eventData);
};


let createRequestXMLDocument = function (requestName, attrMap) {
	let parser = new DOMParser();
	// Create XCIRequest
	let xmlContent = "<?xml version=\'1.0\' encoding=\'UTF-8\' ?><XCI><Request></Request></XCI>";
	let xmlDocument = parser.parseFromString(xmlContent, "application/xml");
	let reqElem = xmlDocument.getElementsByTagName('Request')[0];
	reqElem.setAttribute(ID, requestName);
	reqElem.setAttribute(NAME, requestName);

	attrMap.forEach((value, key, map) => {
		let newElement = xmlDocument.createElement(ATTRIBUTE);
		// newElement.setAttribute(NAME, _.escape(key));
		newElement.setAttribute(NAME, key);
		// newElement.setAttribute(VALUE, _.escape(value));
		newElement.setAttribute(VALUE, value);
		reqElem.appendChild(newElement);
	});

	return xmlDocument;
};

let getPreferenceValueOrDefault = async (prefName, defaultValue) => {
	let prefValue = defaultValue;
	let resPromise = soaSvc.postUnchecked('Administration-2012-09-PreferenceManagement', 'getPreferences', {
		preferenceNames: [prefName],
		includePreferenceDescriptions: false
	});

	let result = await resPromise;
	if (result && result.response && result.response.length > 0) {
		if (result.response[0].values.values.length > 0 && result.response[0].values.values[0] !== '') {
			prefValue = result.response[0].values.values[0];
		}
	}
	return prefValue;

};



/**
* This function gets the rules from the response and returns an array.
*/
export let getRules = function (response, data, ctx) {
	let outputRules = [];
	let count = 0;

	//if (response.ServiceData.modelObjects != undefined) {
	if (response && response.ServiceData && response.ServiceData.modelObjects) {
		// get length of model objects in the response
		let length = Object.keys(response.ServiceData.modelObjects).length;
		for (let i = 0; i < length; i++) {
			let uid = Object.keys(response.ServiceData.modelObjects)[i];
			let displayValue = response.ServiceData.modelObjects[uid].props.object_string.uiValues[0];

			// the the listbox widget the following properties for an element are used:
			// propDisplayValue: Display Name of the element
			// propDisplayDescription: Displayed description of the element
			// dispValue:
			// propInternalValue: Internal value corresponding to the propDisplayValue
			count++;
			outputRules.push({
				propInternalValue: uid,
				propDisplayValue: displayValue
			});

		}
	}

	return {
		rules: outputRules,
		count: count
	};

	// return outputRules;
};

/**
* This function gets the selected item revisions and returns an array.
*/
export let getItemRevisions = function () {
	let outputItemRevisions = [];

	// get selected item revisions
	let items = appCtxSvc.ctx.mselected;
	for (let i = 0; i < items.length; i++) {
		let assignedItemRevision = {};

		// get uid and type to add them to the return array
		assignedItemRevision["uid"] = items[i].uid;
		assignedItemRevision["type"] = items[i].type;
		outputItemRevisions.push(assignedItemRevision);
	}
	return outputItemRevisions;
};

/**
* This function gets the uid for a given object.
*/
export let getUid = function (db, list, revision) {
	let outputUid = undefined;
	for (let i = 0; i < revision.length; i++) {
		if (list.uiValue == revision[i].propDisplayValue) {
			outputUid = revision[i].uid;
		}
	}
	return outputUid;
};

/**
* This function gets the value for the selected expand level.
* "all" - "-1"
* "none" - "0"
* "selectDepth" -  integer
*/
export let getExpand = function (data) {
	let depth = ["", ""];

	if (data.sendTo3DExp_expDepth.dbValue == "all") {
		depth[0] = "all";
		depth[1] = -1;
	}
	if (data.sendTo3DExp_expDepth.dbValue == "none") {
		depth[0] = "none";
		depth[1] = 0;
	}
	if (data.sendTo3DExp_expDepth.dbValue == "selectDepth") {
		depth[0] = "selectDepth";
		let isnum = /^\d+$/.test(data.sendTo3DExp_depth.dbValue);
		if (!isnum || data.sendTo3DExp_depth.dbValue == "") {
			return -1;
		}
		depth[1] = data.sendTo3DExp_depth.dbValue;
	}
	return depth;
};


/**
 * Avinash Desai Implementations start here
 */


/**
 * Send a Request to TCI
 *
 * @param {Document} requestXmlDocument - Selected Object
 * @param {boolean} showAlert - If this function should show alerts, default is false
 * @param {boolean} closePanel - Decision to close Panel after successful request, default is false
 * 
 */
export let makeRequestToTCI = function (requestXmlDocument, showAlert = false, closePanel = false) {

	return new Promise((resolve, reject) => {

		let tciServerUrl = null;
		let responseXmlDocument = null;

		// get Teamcenter preference XCI_SERVER_URL
		soaSvc.postUnchecked('Administration-2012-09-PreferenceManagement', 'getPreferences', {
			preferenceNames: ['XCI_SERVER_URL'],
			includePreferenceDescriptions: false
		}).then(
			function (result) {
				if (result) {
					if (result.partialErrors || result.PartialErrors) {
						if (showAlert) {
							messagingSvc.showWarning("partialErrors" + result.partialErrors + " " + result.PartialErrors);
						}
					}

					// get the Teamcenter preference of the result
					if (result.response[0] != undefined) {
						if (result.response[0].values.values.length > 0 && result.response[0].values.values[0] !== '') {
							tciServerUrl = result.response[0].values.values[0];
						}
					}
				}

				// check if the Teamcenter preference is filled
				if (tciServerUrl == null || tciServerUrl == "") {
					let msg = "Please set Teamcenter Preference XCI_SERVER_URL.";
					if (showAlert) {
						messagingSvc.showWarning(msg);
					}
					reject(msg);
				}

				// create XMLHttpRequest
				let http = new XMLHttpRequest();

				http.open("POST", tciServerUrl + "/XCIRequest", true);

				http.onload = function () {
				};

				http.onerror = function () {
					// write to html console
					let message = "Error:" + "Could not send request to " + tciServerUrl;
					if (showAlert) {
						messagingSvc.showError(message);
					}
					reject(message);
				};

				http.onreadystatechange = function () {
					if (http.readyState == 4 && http.status == 200) {
						responseXmlDocument = http.responseXML;

						//for debug
						// const serializer = new XMLSerializer();
						// const xmlStr = serializer.serializeToString(responseXmlDocument);
						// console.info(xmlStr);

						let msgElements = responseXmlDocument.getElementsByTagName(MESSAGE);
						let messages = [];
						for (const element of msgElements) {
							messages.push(element.childNodes[0].nodeValue);
						}

						if (messages.length > 0) {
							if (showAlert) {
								messagingSvc.showInfo(messages.join());
							}
						}

						// parse xml response and check for error tags
						let errors = responseXmlDocument.getElementsByTagName(ERROR);
						if (errors.length > 0) {
							// show error dialog
							let statusMessages = [];
							for (let x = 0; x < errors.length; x++) {
								// for the value
								statusMessages.push(errors.item(x).innerHTML + " [" + errors.item(x).getAttribute(CODE) + "]");
							}
							let displayResponse = statusMessages.join(', \r\n');
							if (showAlert) {
								messagingSvc.showError('An error occured:\n' + displayResponse);
							}
							reject(displayResponse);
						}

						resolve(responseXmlDocument);
					}
				};

				http.setRequestHeader('Content-type', 'application/xml; charset=utf-8');

				//for debug
				// const serializer = new XMLSerializer();
				// const xmlStr = serializer.serializeToString(inputXmlDocument);
				// console.info(xmlStr);

				try {
					// send xmlDocument
					http.send(requestXmlDocument);

					//To close the panel
					if (closePanel) {
						closeCurrentPanel();
					}

				} catch (e) {
					// Network error, etc.
					let msg = http.response + "--Failed to send Request: " + e;
					if (showAlert) {
						messagingSvc.showError(msg);
					}
					reject(msg);
				}
			},
			function (reason) {
				messagingSvc.showError("Failed to get Teamcenter Preference XCI_SERVER_URL. " + reason.toString());
				tciServerUrl = null;
			}
		);
	});
}


/**
* This function sends Folder request to 3DX.
*/
export let sendFolderTo3DX = function (data) {
	var currentuser = appCtxSvc.getCtx('user');

	let parameters = new Map();
	parameters.set("__SENDER", "TC-AWS");
	parameters.set("__TYPE", "FOLDER_PDM");
	parameters.set("__DESCRIPTION", data.txtObjectString.dbValue);
	parameters.set("OBJECT_STRING", data.txtObjectString.dbValue);
	parameters.set("USER", currentuser.props.user_name.dbValue);
	parameters.set("UID", data.selectedObjectUid);
	parameters.set("TYPE", data.selectedObjectType);
	parameters.set("REVISION", "");
	parameters.set("NAME", data.selectedObjectName);
	parameters.set("REMOVE_EXISTING_NOT_PROVIDED", data.removeNonExistingChildren.dbValue);
	parameters.set("REVISION_RULE", data.useRevisionRule.dbValue ? data.listRevisionRules.uiValue : "");
	parameters.set("REVISION_RULE_UID", data.useRevisionRule.dbValue ? data.listRevisionRules.dbValue : "");
	parameters.set("SEND_CONFIGURATION", false);
	parameters.set("MUST_EXPAND_FLAG", true);
	parameters.set("FOLDER_ONLY", data.onlyFolderStructure.dbValue);
	parameters.set("FORCE", data.force.dbValue);

	let requestXmlDocument = createRequestXMLDocument(START_JOB, parameters);

	makeRequestToTCI(requestXmlDocument, false, false).then((responseXml) => {
		//for debug
		const serializer = new XMLSerializer();
		const xmlStr = serializer.serializeToString(responseXml);
		console.info(xmlStr);

		let msgElements = responseXml.getElementsByTagName(MESSAGE);
		if (msgElements.length > 0) {
			//Collecte all attribute element names and values
			for (let x = 0; x < msgElements.length; x++) {
				let value = msgElements[x].getAttribute('code');
				if (value == '101130') {
					messagingSvc.showInfo("Request successfully created for object: " + data.txtObjectString.dbValue);
					break;
				}
			}
		}

	}).catch((errMsg) => {
		messagingSvc.showError(errMsg);
	});

}


/**
* This function sends the Item Revision request to 3DX
*/
export let sendTo3DX = function (data) {
	let tciServerUrl = null;

	// get values from data
	let objectString = data.sendTo3DExp_txtObjectString.dbValue;
	let userName = data.sendTo3DExp_txtUserName.dbValue;
	let uid = data.sendTo3DExp_txtUid.dbValue;
	let removeNonExistingChildren = data.sendTo3DExp_removeNonExistingChildren.dbValue;
	let transferLocationOfControl = data.sendTo3DExp_transferLocationOfControl.dbValue;
	let transferStructureFilterOnly = data.sendTo3DExp_transferStructureFilterOnly.dbValue;
	let transferSubstructureParallel = data.sendTo3DExp_transferSubstructureParallel.dbValue;
	let includeVariantConditions = data.sendTo3DExp_includeVariantConditions.dbValue;
	let force = data.sendTo3DExp_force.dbValue;
	let isBomActive = data.isBomActive;

	// get checked expand depth
	let expand_depth = exports.getExpand(data);
	if (expand_depth == -1) {
		messagingSvc.showWarning("Only Digits in expand depth!");
		return;
	}

	// get revision rule display name and uid
	let revisionRule = data.sendTo3DExp_useRevision.dbValue ? data.sendTo3DExp_listRevisionRules.uiValue : "";
	let revisionRuleUid = data.sendTo3DExp_useRevision.dbValue ? data.sendTo3DExp_listRevisionRules.dbValue : "";

	// if useRevision is checked, a revision rule is required
	if (data.sendTo3DExp_useRevision.dbValue && (revisionRule == null || revisionRule == "" || revisionRuleUid == null || revisionRuleUid == "undefined")) {
		messagingSvc.showWarning("To use a Revision for expand, a Revision Rule is required.");
		return;
	}
	if (revisionRuleUid != "" && revisionRuleUid != undefined && data.sendTo3DExp_useRevision.dbValue == false) {
		messagingSvc.showWarning("When you have defined a Revision Rule, please check 'Use revision for expand' or remove the Revision Rule.");
		return;
	}
	// prevent the usage of undefined for the uid
	if (data.sendTo3DExp_useRevision.dbValue == false) {
		revisionRuleUid = "";
	}

	// get variant rule display name and uid
	let variantRule = data.sendTo3DExp_useVariant.dbValue ? data.sendTo3DExp_listVariantRules.uiValue : "";
	let variantRuleUid = data.sendTo3DExp_useVariant.dbValue ? data.sendTo3DExp_listVariantRules.dbValue : "";

	// if useVariant is checked, a variant rule is required
	if (data.sendTo3DExp_useVariant.dbValue && (variantRule == null || variantRule == "" || variantRuleUid == null || variantRuleUid == "undefined")) {
		messagingSvc.showWarning("To use a Variant for expand, a Variant Rule is required.");
		return;
	}
	if (variantRuleUid != "" && variantRuleUid != undefined && data.sendTo3DExp_useVariant.dbValue == false) {
		messagingSvc.showWarning("When you have defined a Variant Rule, please check 'Use variant for expand' or remove the Variant Rule.");
		return;
	}

	// prevent the usage of undefined for the uid
	if (data.sendTo3DExp_useVariant.dbValue == false) {
		variantRuleUid = "";
	}

	if (transferSubstructureParallel && expand_depth && (expand_depth[0] != "selectDepth")) {
		transferSubstructureParallel = false;
	}

	// get structure filter name, else empty string
	let structureFilterName = data.sendTo3DExp_structureFilterName.dbValue;
	// if setNameForStructureFilter is checked, a structure filter name is required
	if (data.sendTo3DExp_setNameForStructureFilter.dbValue && (structureFilterName == null || structureFilterName == "")) {
		messagingSvc.showWarning("To transfer a filter, a name for the structure filter is required.");
		return;
	}
	if (structureFilterName != "" && structureFilterName != undefined && data.sendTo3DExp_setNameForStructureFilter.dbValue == false) {
		messagingSvc.showWarning("When you have defined a Structure Filter Name, please check 'Set name for structure filter' or remove the Structure Filter Name.");
		return;
	}

	// The TcSoaService is an asynchronous call.
	// The implementation has to be scheduled by ".then".
	// p.then(onFulfilled, onRejected);
	//
	// p.then(function(value){
	//		// fulfillment
	// }, function(reason){
	//		// rejection
	// });
	//
	// When the Teamcenter Preference is fetched, the flow can be continued. The variable tciServerUrl can be used.

	// get Teamcenter preference XCI_SERVER_URL
	soaSvc.postUnchecked('Administration-2012-09-PreferenceManagement', 'getPreferences', {
		preferenceNames: ['XCI_SERVER_URL'],
		includePreferenceDescriptions: false
	}).then(
		function (result) {
			if (result) {
				if (result.partialErrors || result.PartialErrors) {
					messagingSvc.showWarning("partialErrors" + result.partialErrors + " " + result.PartialErrors);
				}

				// get the Teamcenter preference of the result
				if (result.response[0] != undefined) {
					if (result.response[0].values.values.length > 0 && result.response[0].values.values[0] !== '') {
						tciServerUrl = result.response[0].values.values[0];
					}
				}
			}

			// check if the Teamcenter preference is filled
			if (tciServerUrl == null || tciServerUrl == "") {
				messagingSvc.showError("Please set Teamcenter Preference XCI_SERVER_URL.");
				return;
			}

			//_.escape() is used to resolve the special characters issue

			// Create XCIRequest
			let xmlContent = '<?xml version=\'1.0\' encoding=\'UTF-8\' ?>'
				+ '<XCI>'
				+ '<Request name = "StartJob" id = "StartJob">'
				+ '<Attribute name = "__SENDER" value = "TC-AWS"/>'
				+ '<Attribute name = "__TYPE" value = "OBJECT_PDM"/>'
				+ '<Attribute name = "__DESCRIPTION" value = "' + _.escape(objectString) + '"/>'
				+ '<Attribute name = "OBJECT_STRING" value = "' + _.escape(objectString) + '"/>'
				+ '<Attribute name = "USER" value = "' + _.escape(userName) + '"/>'
				+ '<Attribute name = "UID" value = "' + _.escape(uid) + '"/>'
				+ '<Attribute name = "EXPAND_LEVEL" value = "' + expand_depth[1] + '"/>'
				+ '<Attribute name = "REMOVE_EXISTING_NOT_PROVIDED" value = "' + removeNonExistingChildren + '"/>'
				+ '<Attribute name = "TRANSFER_LOC" value = "' + transferLocationOfControl + '"/>'
				+ '<Attribute name = "TRANSFER_STRUCTURE_FILTER_ONLY" value = "' + transferStructureFilterOnly + '"/>'
				+ '<Attribute name = "REVISION_RULE" value = "' + _.escape(revisionRule) + '"/>'
				+ '<Attribute name = "REVISION_RULE_UID" value="' + _.escape(revisionRuleUid) + '" />'
				+ '<Attribute name = "VARIANT_RULE" value = "' + _.escape(variantRule) + '"/>'
				+ '<Attribute name = "VARIANT_RULE_UID" value="' + _.escape(variantRuleUid) + '" />'
				+ '<Attribute name = "STRUCTURE_FILTER" value = "' + _.escape(structureFilterName) + '"/>'
				+ '<Attribute name = "MUST_EXPAND_FLAG" value="true" />'
				+ '<Attribute name = "FORCE" value = "' + force + '"/>'
				+ '<Attribute name = "TRANSFER_SUBSTRUCTURES_PARALLEL" value = "' + transferSubstructureParallel + '"/>'
				+ '<Attribute name = "SEND_CONFIGURATION" value = "' + includeVariantConditions + '"/>'
				+ '</Request></XCI>';

			// create XMLHttpRequest
			let http = new XMLHttpRequest();

			//TODO: normalize xml befor sending

			http.open("POST", tciServerUrl + "/XCIRequest", true);

			http.onload = function () {
			};

			http.onerror = function () {
				// write to html console
				let message = "Error:" + "Could not send request to " + tciServerUrl;
				messagingSvc.showError(message);
			};

			http.onreadystatechange = function () {
				if (http.readyState == 4 && http.status == 200) {
					let xmlDoc = http.responseXML;

					//<XCI>
					//  <Response id="SendItemRevisionToV6" name="SendItemRevisionToV6">
					//		<Error code="1032">send item revision failed</Error>
					//  </Response>
					//</XCI>

					// parse xml response and check for error tags
					let errors = xmlDoc.getElementsByTagName(ERROR);
					if (errors.length > 0) {
						// show error dialog
						let statusMessages = [];
						for (let x = 0; x < errors.length; x++) {
							// for the value
							statusMessages.push(errors.item(x).innerHTML + " [" + errors.item(x).getAttribute("code") + "]");
						}
						let displayResponse = statusMessages.join(', ');
						messagingSvc.showError("An error occured:\n" + displayResponse);
					} else {
						// show success dialog
						messagingSvc.showInfo("Request successfully created for object: " + objectString);
					}
				}
			};

			http.setRequestHeader('Content-type', 'application/xml; charset=utf-8');

			// parse request string into XML
			let parser = new DOMParser();

			let xmlDocument = parser.parseFromString(xmlContent, "application/xml");

			try {
				// send xmlDocument
				http.send(xmlDocument);

				//To close the panel
				closeCurrentPanel();
			} catch (e) {
				// Network error, etc.
				messagingSvc.showError(http.response + "--Failed to send Request: " + e);
				return;
			}
		},
		function (reason) {
			messagingSvc.showError("Failed to get Teamcenter Preference XCI_SERVER_URL. " + reason.toString());
			tciServerUrl = null;
		}
	);
};


/**
 * Ensures object string property loaded
 *
 * @param {String[]} uidsToLoad - array of uids to load
 * @return {Promise} A promise is return which resolves after 'object_string' properties are loaded
 */
export let ensureObjectString = function (uidsToLoad) {
	// return dms.loadObjects(uidsToLoad).then(function () {
	// 	return dms.getProperties(uidsToLoad, ['object_string']);
	// });
};

/**
 * Adapt and update context, then activate the command panel
 *
 * @param {Object} selectedObj - selected objects
 * @param {Array} selectedObj - array of selected objects
 * @param {String} commandId - ID of the command to open.
 * @param {String} location - Which panel to open the command in
 * @param {Boolean} push - Optional parameter to push workarea content when opening command panel
 * @param {Boolean} closeWhenCommandHidden - Optional parameter to disable the automatic closing of the panel when a command is hidden. Defaults to true.
 * @param {Object} config - Optional parameter to override the configuration attributes of sidenav, which includes width, height and slide.
 */
export let updateSelectionAndActivateSendTo3DXPanel = function (input) {
	let ctx = input.ctx;
	let selectedObj = input.selectedObj;
	let mselected = input.mselected;
	let commandId = input.commandId;
	let location = input.location;
	let push = input.push;
	let config = input.config;
	let closeWhenCommandHidden = input.closeWhenCommandHidden;
	let isBomActive = (selectedObj.type == 'Awb0DesignElement');
	let isSingleRevSelected = (mselected && mselected.length === 1 && selectedObj.modelType.typeHierarchyArray.includes('ItemRevision'));
	let rules = getDefaultRevNVarRules(ctx);
	let selectedObjs = [];
	selectedObjs.push(selectedObj);
	let adaptedObjsPromise = adapterSvc.getAdaptedObjects(selectedObjs);
	adaptedObjsPromise.then(function (adaptedObjs) {

		if (adaptedObjs.length > 0) {
			selectedObj = adaptedObjs[0];
		}

		dms.loadObjects([selectedObj.uid]).then(function () {
			let context;
			context = {
				selectedObjects: [selectedObj],
				isBomActive: isBomActive,
				isSingleRevSelected: isSingleRevSelected,
				defaultRules: rules
			};

			if (commandId) { commandPanelService.activateCommandPanel(commandId, location, context, push, closeWhenCommandHidden, config); }

		});

	});

};

const createPropertyInfo = (propertyName, propertyDisplayName, propertyValue, propertyDisplayValue, isEnabled) => {
	return {
		type: 'STRING',
		hasLov: false,
		isArray: false,
		displayValue: propertyDisplayValue,
		uiValue: propertyDisplayValue,
		value: propertyValue,
		dbValue: propertyValue,
		propertyName: propertyName,
		propertyDisplayName: propertyDisplayName,
		isEnabled: isEnabled
	};
};

const createColumnInfo = (name, displayName, widthVal, enableColumnMenu) => {
	return {
		name: name,
		displayName: displayName,
		minWidth: 20,
		width: widthVal,
		/* enableColumnMoving: true, */
		enableColumnMenu: enableColumnMenu,
		isFilteringEnabled: true,
		enableSorting: true,
		enableColumnResizing: true
	};
};


/**
 * Fetch Process Status Data from TCI
 *
 * @param {Object} user - Current User
 * @param {Object} ctx - Cuttent Context
 * @param {String} commandId - ID of the command to open.
 * @param {String} location - Which panel to open the command in
 * @param {Boolean} push - Optional parameter to push workarea content when opening command panel
 * @param {Boolean} closeWhenCommandHidden - Optional parameter to disable the automatic closing of the panel when a command is hidden. Defaults to true.
 * @param {Object} config - Optional parameter to override the configuration attributes of sidenav, which includes width, height and slide.
 */
export let fetchProcessDataFromTci = function (input) {
	let user = input.ctx.user;
	let commandId = input.commandId;
	let location = input.location;
	let push = input.push;
	let config = input.config;
	let closeWhenCommandHidden = input.closeWhenCommandHidden;

	let tciProcesses = [];
	let totalTciProcessesCount = 0;
	let columnsInfo = [];

	let parameters = new Map();
	parameters.set("USER", user.props.user_name.dbValue);

	let requestXmlDocument = createRequestXMLDocument(GET_PROCESS_INFO_FOR_USER, parameters);

	makeRequestToTCI(requestXmlDocument, false).then((responseXml) => {
		let columnNames = [];
		let pds = responseXml.getElementsByTagName("PD");
		if (pds.length > 0) {
			for (let x = 0; x < pds.length; x++) {
				++totalTciProcessesCount;

				let tciProcess = {
					type: "TciProcess",
					uid: totalTciProcessesCount,
					props: {}
				};

				let props = {};

				if (pds[x].hasAttributes()) {
					for (const attr of pds[x].attributes) {
						let prop = createPropertyInfo(attr.name, attr.name, attr.value, attr.value, true);

						if (!columnNames.includes(attr.name)) {
							columnNames.push(attr.name);
						}

						props[prop.propertyName] = prop;
					}
				}

				tciProcess.props = props;
				tciProcesses.push(tciProcess);
			}

			//Before the Column sequence was as it was coming form teh XML. Now below new implementation will change the sequence.
			// columnNames.forEach(cn => {

			// 	let ci = createColumnInfo(cn, cn, 150, true);
			// 	columnsInfo.push(ci);
			// });

			// New Column Config start
			let attrName = 'Activity';
			if (columnNames.includes(attrName)) {
				let ci = createColumnInfo(attrName, attrName, 70, true);
				columnsInfo.push(ci);
			}
			attrName = 'Status';
			if (columnNames.includes(attrName)) {
				let ci = createColumnInfo(attrName, attrName, 70, true);
				columnsInfo.push(ci);
			}
			attrName = 'Description';
			if (columnNames.includes(attrName)) {
				let ci = createColumnInfo(attrName, attrName, 320, true);
				columnsInfo.push(ci);
			}
			attrName = 'Start';
			if (columnNames.includes(attrName)) {
				let ci = createColumnInfo(attrName, attrName, 150, true);
				columnsInfo.push(ci);
			}
			attrName = 'End';
			if (columnNames.includes(attrName)) {
				let ci = createColumnInfo(attrName, attrName, 150, true);
				columnsInfo.push(ci);
			}
			attrName = 'User';
			if (columnNames.includes(attrName)) {
				let ci = createColumnInfo(attrName, attrName, 100, true);
				columnsInfo.push(ci);
			}
			// New Column Config End


			let context;
			context = {
				tciProcesses: tciProcesses,
				totalTciProcessesCount: totalTciProcessesCount,
				columnsInfo: columnsInfo
			};

			if (commandId) { commandPanelService.activateCommandPanel(commandId, location, context, push, closeWhenCommandHidden, config); }

		} else {
			messagingSvc.showWarning('No Process Elements found for current User.');
			return;
		}

	}).catch((errMsg) => {
		closeCurrentPanel();
		messagingSvc.showError(errMsg);
	});
};


export let loadTciProcessColumns = (columnsInfo, dataProvider) => {
	dataProvider.columnConfig = {
		columns: columnsInfo
	};

	return columnsInfo;
};

/**
 * This function is responsible for handling pagination of the data for Process Status Table.
 *
 * @param {ObjectArray} response Response from GET call
 * @param {int} startIndex Start index for pagination
 * @param {int} pageSize Total number of rows to return for pagination
 *
 * @return {ObjectArray} Paginated filtered result
 */
export const loadTciProcessData = function (tciProcesses, startIndex, pageSize, sortCriteria) {
	if (tciProcesses) {
		let tempTciProcesses = tciProcesses;

		if (sortCriteria && sortCriteria.length > 0) {
			let criteria = sortCriteria[0];
			let sortDirection = criteria.sortDirection;
			let sortColName = criteria.fieldName;

			if (sortDirection === 'ASC') {
				tempTciProcesses.sort(function (a, b) {
					if (a.props[sortColName].value <= b.props[sortColName].value) {
						return -1;
					}
					return 1;
				});
			} else if (sortDirection === 'DESC') {
				tempTciProcesses.sort(function (a, b) {
					if (a.props[sortColName].value >= b.props[sortColName].value) {
						return -1;
					}
					return 1;
				});
			}
		}

		let endIndex = startIndex + pageSize;
		let searchResults = tempTciProcesses.slice(startIndex, endIndex);
		return searchResults;
	}
	return [];
};


/**
 * Adapt and update context, then activate the command panel
 *
 * @param {Object} ctx - Cuttent Context
 * @param {String} commandId - ID of the command to open.
 * @param {String} location - Which panel to open the command in
 * @param {Boolean} push - Optional parameter to push workarea content when opening command panel
 * @param {Boolean} closeWhenCommandHidden - Optional parameter to disable the automatic closing of the panel when a command is hidden. Defaults to true.
 * @param {Object} config - Optional parameter to override the configuration attributes of sidenav, which includes width, height and slide.
 */
export let fetchPropsAndActivateViewPropsFrom3DExpPanel = function (input) {
	let tciServerUrl = null;
	let ctx = input.ctx;
	let commandId = input.commandId;
	let location = input.location;
	let push = input.push;
	let config = input.config;
	let closeWhenCommandHidden = input.closeWhenCommandHidden;
	let isBomActive = (ctx.selected.type == 'Awb0DesignElement');
	let isSingleRevSelected = (ctx.mselected && ctx.mselected.length === 1 && ctx.selected.modelType.typeHierarchyArray.includes('ItemRevision'));
	let selectedObj = ctx.selected;
	let uid = selectedObj.uid;

	if (isBomActive != true && isSingleRevSelected != true) {
		messagingSvc.showWarning("Please select a single Item Revision to Update the Properties in 3DEXPERIENCE!");
		return;
	}

	let selectedObjs = [];
	selectedObjs.push(selectedObj);
	let adaptedObjsPromise = adapterSvc.getAdaptedObjects(selectedObjs);

	adaptedObjsPromise.then(function (adaptedObjs) {

		if (adaptedObjs.length > 0) {
			selectedObj = adaptedObjs[0];
		}

		uid = selectedObj.uid;
		ensureObjectString([uid]);

		let parameters = getCredentialsMap(expData.loggedUser.user.dbValue, expData.loggedUser.secret.dbValue, expData.loggedUser.role.dbValue, expData.loggedUser.vault.dbValue, expData.loggedUser.isPasswordEncrypted);
		parameters.set(TC_REV_OBID, uid);
		let requestXmlDocument = createRequestXMLDocument(GET_ITEM_REVISION_PROPERTIES, parameters);

		makeRequestToTCI(requestXmlDocument, false).then((responseXml) => {

			let attrs = responseXml.getElementsByTagName(ATTRIBUTE);
			if (attrs.length > 0) {
				let propArray = [];
				for (let x = 0; x < attrs.length; x++) {
					if (attrs[x].hasAttributes()) {
						let prop = {
							name: "",
							displayName: "",
							type: 'STRING',
							isEditable: false,
							isRequired: false,
							dbValue: "",
							dispValue: "",
							labelPosition: 'PROPERTY_LABEL_AT_SIDE'
						};

						for (const attr of attrs[x].attributes) {
							switch (attr.name) {
								case NAME:
									prop['name'] = attr.value;
									break;

								case DISPLAY:
									prop['displayName'] = attr.value;
									break;

								case VALUE:
									prop['dbValue'] = attr.value;
									prop['dispValue'] = attr.value;
									break;

							}

						}
						propArray.push(prop);
					}
				}


				let context;
				context = {
					selectedObject: selectedObj,
					propArray: propArray
				};

				if (commandId) { commandPanelService.activateCommandPanel(commandId, location, context, push, closeWhenCommandHidden, config); }

			} else {
				messagingSvc.showWarning('No properties available for the selected product.');
				return;
			}

		}).catch((errMsg) => {
			messagingSvc.showError(errMsg);
		});

	}).catch((errMsg) => {
		messagingSvc.showError(errMsg);
	});
};

/**
 * Adapt and update context, then activate the command panel
 *
 * @param {Object} ctx - Cuttent Context
 * @param {String} commandId - ID of the command to open.
 * @param {String} location - Which panel to open the command in
 * @param {Boolean} push - Optional parameter to push workarea content when opening command panel
 * @param {Boolean} closeWhenCommandHidden - Optional parameter to disable the automatic closing of the panel when a command is hidden. Defaults to true.
 * @param {Object} config - Optional parameter to override the configuration attributes of sidenav, which includes width, height and slide.
 */
export let fetchPropInfoAndActivateUpdatePropsIn3DXPanel = function (input) {
	let tciServerUrl = null;
	let ctx = input.ctx;
	let commandId = input.commandId;
	let location = input.location;
	let push = input.push;
	let config = input.config;
	let closeWhenCommandHidden = input.closeWhenCommandHidden;
	let isBomActive = (ctx.selected.type == 'Awb0DesignElement');
	let isSingleRevSelected = (ctx.mselected && ctx.mselected.length === 1 && ctx.selected.modelType.typeHierarchyArray.includes('ItemRevision'));
	let selectedObj = ctx.selected;
	let uid = selectedObj.uid;

	if (isBomActive != true && isSingleRevSelected != true) {
		messagingSvc.showWarning("Please select a single Item Revision to Update the Properties in 3DEXPERIENCE!");
		return;
	}

	let selectedObjs = [];
	selectedObjs.push(selectedObj);
	let adaptedObjsPromise = adapterSvc.getAdaptedObjects(selectedObjs);
	adaptedObjsPromise.then(function (adaptedObjs) {

		if (adaptedObjs.length > 0) {
			selectedObj = adaptedObjs[0];
		}

		uid = selectedObj.uid;
		ensureObjectString([uid]);


		// get Teamcenter preference XCI_SERVER_URL
		soaSvc.postUnchecked('Administration-2012-09-PreferenceManagement', 'getPreferences', {
			preferenceNames: ['XCI_SERVER_URL'],
			includePreferenceDescriptions: false
		}).then(
			function (result) {
				if (result) {
					if (result.partialErrors || result.PartialErrors) {
						messagingSvc.showWarning("partialErrors" + result.partialErrors + " " + result.PartialErrors);
					}

					// get the Teamcenter preference of the result
					if (result.response[0] != undefined) {
						if (result.response[0].values.values.length > 0 && result.response[0].values.values[0] !== '') {
							tciServerUrl = result.response[0].values.values[0];
						}
					}
				}

				// check if the Teamcenter preference is filled
				if (tciServerUrl == null || tciServerUrl == "") {
					messagingSvc.showError("Please set Teamcenter Preference XCI_SERVER_URL.");
					return;
				}

				//_.escape() is used to resolve the special characters issue

				// let xd = createRequestXMLDocument(GET_META_UPDATE_INFO_FOR_UID_REQUEST_NAME, new Map([['UID', uid]]));
				// const serializer = new XMLSerializer();
				// const xmlStr = serializer.serializeToString(xd);
				// create XMLHttpRequest
				let http = new XMLHttpRequest();

				//TODO: normalize xml befor sending

				http.open("POST", tciServerUrl + "/XCIRequest", true);

				http.onload = function () {
				};

				http.onerror = function () {
					// write to html console
					let message = "Error:" + "Could not send request to " + tciServerUrl;
					messagingSvc.showError(message);
				};

				http.onreadystatechange = function () {
					if (http.readyState == 4 && http.status == 200) {
						let xmlDoc = http.responseXML;

						// parse xml response and check for error tags
						let errors = xmlDoc.getElementsByTagName(ERROR);
						if (errors.length > 0) {
							// show error dialog
							let statusMessages = [];
							for (let x = 0; x < errors.length; x++) {
								// for the value
								statusMessages.push(errors.item(x).innerHTML + " [" + errors.item(x).getAttribute(CODE) + "]");
							}
							let displayResponse = statusMessages.join(', ');
							messagingSvc.showError('An error occured:\n' + displayResponse);
							return;
						}

						let msgs = xmlDoc.getElementsByTagName(MESSAGE);
						if (msgs.length > 0) {
							let code = msgs[0].getAttribute(CODE);
							if (code == '101126') {
								messagingSvc.showWarning('No properties for Update found.\nPlease check your configuration!');
								return;
							}
						}

						let muis = xmlDoc.getElementsByTagName(MUI);
						if (muis.length > 0) {
							let propArray = [];
							for (let x = 0; x < muis.length; x++) {
								if (muis[x].hasAttributes()) {
									let propObject = {};
									for (const attr of muis[x].attributes) {
										propObject[attr.name] = attr.value;
									}
									propArray.push(propObject);
								}
							}


							let context;
							context = {
								selectedObject: selectedObj,
								propArray: propArray
							};

							if (commandId) { commandPanelService.activateCommandPanel(commandId, location, context, push, closeWhenCommandHidden, config); }

						} else {
							messagingSvc.showWarning('Something went wrong! No MUI elements found.');
							return;
						}


					}
				};

				http.setRequestHeader('Content-type', 'application/xml; charset=utf-8');

				// parse request string into XML
				// let parser = new DOMParser();

				// let xmlDocument = parser.parseFromString(xmlContent, "application/xml");
				let xmlDocument = createRequestXMLDocument(GET_META_UPDATE_INFO_FOR_UID_REQUEST_NAME, new Map([['UID', uid]]));

				try {
					// send xmlDocument
					http.send(xmlDocument);

					//To close the panel
					//closeCurrentPanel();
				} catch (e) {
					// Network error, etc.
					messagingSvc.showError(http.response + "--Failed to send Request: " + e);
					return;
				}
			},
			function (reason) {
				messagingSvc.showÊrror("Failed to get Teamcenter Preference XCI_SERVER_URL. " + reason.toString());
				tciServerUrl = null;
			}
		);

	});



	return;
}

/**
 * Update data in 3DEXPERIANCE for already exported Object
 *
 * @param {Object} selectedObject - Selected Object
 * @param {Array} vmpArray - Array of vmProperties
 */
export let performUpdatePropsIn3DExp = function (selectedObject, vmpArray, i18n) {
	let tciServerUrl = null;
	// let selectedObject = input.selectedObject;
	// let vmpArray = input.vmpArray;
	let index = 0;

	//Load properties if not loaded
	dms.getProperties([selectedObject.uid], ['object_string', 'item_id', 'item_revision_id']);
	let objString = selectedObject.props.object_string.dbValue;

	//Collect Parameters for Update
	let parameters = new Map();
	parameters.set("__SENDER", "TC-AWC");
	parameters.set("__TYPE", "UPDATE_IN_3DX");
	parameters.set("__DESCRIPTION", objString);
	parameters.set("OBJECT_STRING", objString);

	vmpArray.forEach(vmProp => {
		parameters.set(("ATTR_PDM_NAME_" + index), vmProp.props.pdmName.value);
		parameters.set(("ATTR_VALUE_" + index), vmProp.props.pdmValue.value);
		parameters.set(("ATTR_V6_NAME_" + index), vmProp.props.v6Name.value);
		index++;
	});

	parameters.set("ATTR_NUM", index.toString());

	parameters.set("REVISION", selectedObject.props.item_revision_id.dbValue);
	parameters.set("NAME", selectedObject.props.item_id.dbValue);

	var currentuser = appCtxSvc.getCtx('user');
	parameters.set("USER", currentuser.props.user_name.dbValue);
	parameters.set("UID", selectedObject.uid);
	parameters.set("TYPE", selectedObject.type);

	//Create request XML
	let requestXmlDocument = createRequestXMLDocument(START_JOB, parameters);

	//Make request on TCI
	makeRequestToTCI(requestXmlDocument, false, true).then((responseXml) => {

		//for debug
		// const serializer = new XMLSerializer();
		// const xmlStr = serializer.serializeToString(responseXml);
		// console.info(xmlStr);

		let msgElements = responseXml.getElementsByTagName(MESSAGE);
		if (msgElements.length > 0) {
			//Collecte all attribute element names and values
			for (let x = 0; x < msgElements.length; x++) {
				let value = msgElements[x].getAttribute('code');
				if (value == '101130') {
					messagingSvc.showInfo(i18n.updatePropertiesInitiatedMsg);
					break;
				}
			}
		}

	}).catch((errMsg) => {
		messagingSvc.showError(errMsg);
	});

}

/**
 * check if Website is reachable
 * @param {Website URL} url 
 * @returns Info Message to add after URL
 */
export let checkWebsite = async function (url) {
	let result = await fetch(url, { signal: AbortSignal.timeout(2000) })
		.then(response => {
			if (response.ok) {
				console.log("Website is reachable.");
				return ' OK!';
			} else {
				console.error("Website " + url + " is not reachable. Status code: " + response.status + " " + response.statusText);
				return ' not reachable, ' + response.status + ' ' + response.statusText + '!';
			}
		})
		.catch(error => {
			let errInfo = ' not reachable, 404 Not Found!';
			if (error.name === "TimeoutError") {
				console.error("Timeout: It took more than 2 seconds to get the result!");
				errInfo = ' Timeout, takes longer!';
			} else if (error.name === "AbortError") {
				console.error("Fetch aborted by user action (browser stop button, closing tab, etc.). If you are using Chrome, it may be timeout because it took more than 2 seconds.");
				errInfo = ' Timeout, takes longer!';
			} else if (error instanceof TypeError) {
				console.error("Failed to fetch the website. Check network connectivity or CORS settings.");
			} else {
				console.error("Error reaching the website " + url, error);
			}
			//It will come here in-case the URL is not reachable, so 404 error.
			return errInfo;
		});
	return result;
}

/**
 * Return the XCI Server URL & its Status
 */
export let getXCIServerInfo = async function () {
	let xciUrlInfo = 'Failed to retrieve preference XCI_SERVER_URL!';

	// get Teamcenter preference XCI_SERVER_URL
	await prefSvc.getStringValue('XCI_SERVER_URL').then(
		async function (result) {
			// Check if there is no valid value returned.
			if (!result || result < 0 || result.length < 5) {
				xciUrlInfo = "Failed to retrieve preference XCI_SERVER_URL!";
			}
			else {
				xciUrlInfo = result + await checkWebsite(result + "/XCIRequest");
			}
			return xciUrlInfo;
		}).then((resp) => {
			xciUrlInfo = resp;
		});

	return xciUrlInfo;
};

/**
 * Display the "Help About" Information popup
 */
export let showAboutTCI = async function () {
	//The defer is used because this function is being called with JSFunctionAsync call, so the menu will close immediately.
	var deferred = AwPromiseService.instance.defer();
	deferred.resolve();

	let serverInfo = await getXCIServerInfo();

	let aboutTCIStr = 'TCI-3DEXPERIENCE Teamcenter Integration<br>';
	aboutTCIStr += 'TCI Version (' + TCI_VERSION + ')<br>';
	aboutTCIStr += 'Build Date: (' + TCI_RELEASE_DATE + ')<br>';
	aboutTCIStr += 'TCI Server: ' + serverInfo + '<br>';
	aboutTCIStr += '&#169; 2012, ' + TCI_COPY_YEAR + ' T-Systems International GmbH';
	messagingSvc.showInfo(aboutTCIStr);

	return deferred.promise;
};

/**
 * Login Window related Functions
 */

/**
 * Populate Login Information in Login DIalog of 3DEXPERIENCE
 * @param {Object} data - data Object
 */
export let populate3DExpLoginInfo = async function (data) {

	if (expData) {
		if (expData.loggedUser) {
			if (expData.loggedUser.user) {
				expData.expUser = expData.loggedUser.user;
			}

			expData.expPassword = {
				dbValue: "",
				dispValue: ""
			};

			if (expData.loggedUser.role) {
				expData.expRole = expData.loggedUser.role;
			}

			if (expData.loggedUser.vault) {
				expData.expVault = expData.loggedUser.vault;
			}

			if (expData.loggedUser.connectorType) {
				expData.enoviaConnectorType = expData.loggedUser.connectorType;
			}

		}

		return expData;
	}

	let currentuser = appCtxSvc.getCtx('user');
	let user_name = currentuser.props.user_name;

	expData = {
		expUser: {
			dbValue: user_name.dbValue,
			dispValue: user_name.uiValue
		},
		expPassword: {
			dbValue: "",
			dispValue: ""
		},
		expRole: {},
		expVault: {},
		enoviaProductTypes: [],
		enoviaConnectorType: ""
	};

	let requestXmlDocument = createRequestXMLDocument(GET_ENOVIA_QUERY_SETTINGS, new Map());

	await makeRequestToTCI(requestXmlDocument, false).then((responseXml) => {
		// for debug
		// const serializer = new XMLSerializer();
		// const xmlStr = serializer.serializeToString(responseXml);
		// console.info(xmlStr);

		let attributeElements = responseXml.getElementsByTagName(ATTRIBUTE);
		if (attributeElements.length > 0) {
			let typeDBNamesStr, typeDispNamesStr;

			for (let x = 0; x < attributeElements.length; x++) {
				if (attributeElements[x].hasAttributes()) {

					switch (attributeElements[x].getAttribute(NAME)) {
						case ENOVIA_DEFAULT_ENOVIA_ROLE:
							// expData.expRoleValue = attributeElements[x].getAttribute(VALUE);
							expData.expRole.dbValue = attributeElements[x].getAttribute(VALUE);
							expData.expRole.dispValue = attributeElements[x].getAttribute(VALUE);
							break;

						case ENOVIA_DEFAULT_ENOVIA_VAULT:
							// expData.expVaultValue = attributeElements[x].getAttribute(VALUE);
							expData.expVault.dbValue = attributeElements[x].getAttribute(VALUE);
							expData.expVault.dispValue = attributeElements[x].getAttribute(VALUE);
							break;

						case ENOVIA_QUERY_TYPE_NAME_LIST:
							typeDBNamesStr = attributeElements[x].getAttribute(VALUE);
							break;

						case ENOVIA_QUERY_TYPE_DISPLAY_LIST:
							typeDispNamesStr = attributeElements[x].getAttribute(VALUE);
							break;

						case ENOVIA_CONNECTOR_TYPE:
							expData.enoviaConnectorType = attributeElements[x].getAttribute(VALUE);
							break;
					}
				}
			}

			if (typeDBNamesStr && typeDispNamesStr) {
				const typeDBNames = typeDBNamesStr.split(LIST_SEPARATOR);
				const typeDispNames = typeDispNamesStr.split(LIST_SEPARATOR);

				if (typeDBNames.length > 0 && typeDBNames.length == typeDispNames.length) {
					for (let i = 0; i < typeDBNames.length; ++i) {
						let envType = {
							dbValue: typeDBNames[i],
							dispValue: typeDispNames[i],
							propInternalValue: typeDBNames[i],
							propDisplayValue: typeDispNames[i],
						};

						expData.enoviaProductTypes.push(envType);

					}
				}

			}

		} else {
			// closeCurrentPanel();
			messagingSvc.showWarning("Response XML do not have any Attribute elements.");
		}

	}).catch((errMsg) => {
		closeCurrentPanel();
		messagingSvc.showError(errMsg);
	});
	return expData;
};

export let getCredentialsMap = (userNameValue, userPasswdValue, userRoleValue, userVaultValue, isPasswordEncrypted) => {
	let attMap = new Map();

	attMap.set("user", userNameValue);

	if (isPasswordEncrypted) {
		attMap.set("encryptedPassword", userPasswdValue);
	}
	else {
		// see com.tsystems.xci.xmlservlet.SessionRequest.getEnoviaConnector(Request)
		// for conversion
		/* string to low byte high byte */
		let usPwConv = toHexString(userPasswdValue);
		attMap.set("password", usPwConv);
	}

	attMap.set("role", userRoleValue);
	attMap.set("vault", userVaultValue);

	return attMap;
};


// export let updateLoggedInData = (data) => {
// 	if (data.loggedIn === true && expData.loggedUser) {
// 		return true;
// 	}
// 	return false;
// };

export let logoutOf3DX = () => {
	delete expData.loggedUser.secret;
	if (expData.expPassword) {
		delete expData.expPassword;
	}
	expData.loggedUser.loggedIn = false;
	expData.loggedUser.secret = {};
	expData.loggedUser.isPasswordEncrypted = false;
	expData.expPassword = {};

	//This is required to set the menus visible which depends on the logged in condition
	// appCtxSvc.unRegisterCtx(IS_TCI_LOGGED_IN);
	setTciLoggedOut();

	messagingSvc.showInfo("Logout of 3DEXPERIENCE Successful!");

};


/**
 * Login to 3DEXPERIENCE
 * @param {string} userNameValue - 3DEXPERIENCE Username
 * @param {string} userPasswdValue - 3DEXPERIENCE Password
 * @param {string} userRoleValue - 3DEXPERIENCE Role
 * @param {string} userVaultValue - 3DEXPERIENCE Vault
 */
export let loginTo3DExp = (userNameValue, userPasswdValue, userRoleValue, userVaultValue, i18n) => {
	if (userNameValue && userPasswdValue && userRoleValue && userVaultValue) {
		let parameters = getCredentialsMap(userNameValue.dbValue, userPasswdValue.dbValue, userRoleValue.dbValue, userVaultValue.dbValue, false);
		let requestXmlDocument = createRequestXMLDocument(ENOVIA_LOGIN, parameters);


		makeRequestToTCI(requestXmlDocument, false).then((responseXml) => {

			// ENOVIA login successful; store credentials
			expData.loggedUser = {
				loggedIn: true,
				isPasswordEncrypted: false,
				user: {
					dbValue: userNameValue.dbValue,
					dispValue: userNameValue.dispValue
				},
				secret: {
					dbValue: userPasswdValue.dbValue,
					dispValue: userPasswdValue.dispValue
				},
				role: {
					dbValue: userRoleValue.dbValue,
					dispValue: userRoleValue.dispValue
				},
				vault: {
					dbValue: userVaultValue.dbValue,
					dispValue: userVaultValue.dispValue
				},
				connectorType: expData.enoviaConnectorType
			};

			//for debug
			// const serializer = new XMLSerializer();
			// const xmlStr = serializer.serializeToString(responseXml);
			// console.info(xmlStr);

			let attrElements = responseXml.getElementsByTagName(ATTRIBUTE);
			if (attrElements.length > 0) {
				//Collecte all attribute element names and values
				for (let x = 0; x < attrElements.length; x++) {
					let name = attrElements[x].getAttribute(NAME);
					let value = attrElements[x].getAttribute(VALUE);
					if (name.toLowerCase() == 'encryptedpassword') {
						expData.loggedUser.isPasswordEncrypted = true;
						expData.loggedUser.secret.dbValue = value;
						expData.loggedUser.secret.dispValue = value;
						break;
					}
				}
			}

			delete expData.expPassword;

			//This is required to set the menus visible which depends on the logged in condition
			// appCtxSvc.registerCtx(IS_TCI_LOGGED_IN, true);
			setTciLoggedIn();

			closeCurrentPanel();
			messagingSvc.showInfo("Login to 3DEXPERIENCE Successful!");

		}).catch((errMsg) => {
			// closeCurrentPanel();
			// 	let msgElements = responseXml.getElementsByTagName(MESSAGE);
			// if (msgElements.length > 0) {
			// 	//Collecte all attribute element names and values
			// 	for (let x = 0; x < msgElements.length; x++) {
			// 		let value = msgElements[x].getAttribute('code');
			// 		if (value == '101130') {
			// 			messagingSvc.showInfo(i18n.updatePropertiesInitiatedMsg);
			// 			break;
			// 		}
			// 	}
			// }
			if (errMsg.includes('[1100]')) {
				messagingSvc.showError(i18n.failedToTciLogin);
			} else {
				messagingSvc.showError(errMsg);
			}
		});
	}
};


/**
 * Adapt and update context, then activate the command panel
 *
 * @param {Object} ctx - Cuttent Context
 * @param {String} commandId - ID of the command to open.
 * @param {String} location - Which panel to open the command in
 * @param {Boolean} push - Optional parameter to push workarea content when opening command panel
 * @param {Boolean} closeWhenCommandHidden - Optional parameter to disable the automatic closing of the panel when a command is hidden. Defaults to true.
 * @param {Object} config - Optional parameter to override the configuration attributes of sidenav, which includes width, height and slide.
 */
export let activateQueryProductsIn3DXPanel = function (input) {
	let commandId = input.commandId;
	let location = input.location;
	let push = input.push;
	let config = input.config;
	let closeWhenCommandHidden = input.closeWhenCommandHidden;

	let context;
	context = {
		enoviaProductTypes: expData.enoviaProductTypes
	};

	if (commandId) { commandPanelService.activateCommandPanel(commandId, location, context, push, closeWhenCommandHidden, config); }

};

export const updateFieldWithDefaultValues = (dataList) => {
	let result = {
		type: {
			dbValue: '',
			uiValue: ''
		},
		name: {
			dbValue: '',
			uiValue: ''
		},
		revision: {
			dbValue: '',
			uiValue: ''
		},
		title: {
			dbValue: '',
			uiValue: ''
		}
	};

	if (!expQueryInput) {
		const queryDataJson = localStrg.get('tciData.queryInput');
		expQueryInput = JSON.parse(queryDataJson);
		if (_.isEmpty(expQueryInput)) {
			expQueryInput = undefined;
		}

	}

	if (expQueryInput) {
		result.type.dbValue = expQueryInput.type;
		result.type.uiValue = expQueryInput.type;

		result.name.dbValue = expQueryInput.name;
		result.name.uiValue = expQueryInput.name;

		result.revision.dbValue = expQueryInput.revision;
		result.revision.uiValue = expQueryInput.revision;

		result.title.dbValue = expQueryInput.title;
		result.title.uiValue = expQueryInput.title;

		if (result.type.dbValue != '' && dataList && dataList.length > 0) {
			const elem = dataList.find((el) => el.dbValue == result.type.dbValue);
			if (elem) {
				result.type.uiValue = elem.dispValue;
			}
		}
	}
	else if (dataList && dataList.length > 0) {
		result.type.dbValue = dataList[0].dbValue;
		result.type.uiValue = dataList[0].dispValue;
	}

	return result;
};


async function performQueryIn3DX(queryInput) {
	// let type = queryInput.type;
	// let name = queryInput.name;
	// let revision = queryInput.revision;
	// let title = queryInput.title;
	let envObjs = [];

	let parameters = getCredentialsMap(expData.loggedUser.user.dbValue, expData.loggedUser.secret.dbValue, expData.loggedUser.role.dbValue, expData.loggedUser.vault.dbValue, expData.loggedUser.isPasswordEncrypted);
	parameters.set("type", queryInput.type);
	parameters.set("name", queryInput.name);
	parameters.set("revision", queryInput.revision);
	parameters.set("title", queryInput.title);

	let requestXmlDocument = createRequestXMLDocument(ENOVIA_QUERY, parameters);

	await makeRequestToTCI(requestXmlDocument, false).then((responseXml) => {
		let columnNames = [];
		let attrElements = responseXml.getElementsByTagName(ATTRIBUTE);
		if (attrElements.length > 0) {
			let totalEnvObjectsCount = 0;
			let properties = [];
			let propValues = [];


			//Collecte all attribute element names and values
			for (let x = 0; x < attrElements.length; x++) {
				let attrs = attrElements[x].attributes;
				if (attrs.name && attrs.name.value) {
					if (attrs.name.value == 'attribute[PLMEntity.PLM_ExternalID]') {
						properties.push('PLM_ExternalID');
						propValues.push(attrs.value.value);
					} else {
						properties.push(attrs.name.value);
						propValues.push(attrs.value.value);
					}
				}

			}

			//Get total Objects Count
			totalEnvObjectsCount = propValues[0].split(LIST_SEPARATOR).length;

			//Create base objects for all provided objects with default type VPMReference
			for (let y = 0; y < totalEnvObjectsCount; y++) {
				let baseObject = {
					type: "VPMReference",
					uid: (y + 1).toString(),
					props: {}
				};
				envObjs.push(baseObject);
			}

			//Add the props on all the objects created in above step and populate the props and there values
			for (let z = 0; z < properties.length; z++) {
				let values = propValues[z].split(LIST_SEPARATOR);

				for (let y = 0; y < totalEnvObjectsCount; y++) {
					envObjs[y].props[properties[z]] = values[y];

					if (properties[z] == 'type') {
						envObjs[y].type = values[y];
					}
				}
			}

		} else {
			messagingSvc.showWarning('No Elements found for current Search Criteria.');
		}

	}).catch((errMsg) => {
		closeCurrentPanel();
		messagingSvc.showError(errMsg);
	});

	return envObjs;
}


export let queryIn3DExpAndActivateProductsIn3DXPage = async (input) => {
	let type = input.type.dbValue;
	let name = input.name.dbValue;
	let revision = input.revision.dbValue;
	let title = input.title.dbValue;
	let envObjs = [];

	expQueryInput = {
		type: type,
		name: name,
		revision: revision,
		title: title
	};

	closeCurrentPanel();

	let locContext = appCtxSvc.ctx.locationContext;
	let activeLoc = locContext ? locContext['ActiveWorkspace:SubLocation'] : '';
	if (activeLoc == 'productsIn3DX') {
		localStrg.publish('tciData.queryInput', JSON.stringify(expQueryInput));
		initializeQueriedProductsFrom3DX();
		// reloadPageProductsIn3DX();
	} else {

		envObjs = await performQueryIn3DX(expQueryInput);

		if (envObjs && envObjs.length > 0) {
			// localStrg.publish('expData', JSON.stringify(expData));
			localStrg.publish('tciData.queryInput', JSON.stringify(expQueryInput));
			// localStrg.publish('queryDataFrom3DX', JSON.stringify(envObjs));
			navigateToURL("productsIn3DX");
		}
	}
};


/**
 * Clear Provided Property Value
 * @param {Object} prop - property Object
 */
export let clearProperty = function (prop) {
	prop.dbValue = '';
	prop.uiValue = '';
};

export let getAllOrgForCollabSpace = function (collabSpace) {
	let orgs = [];
	if (collabSpace == undefined) {
		return orgs;
	}
	if (expData && expData.roleTree) {
		expData.roleTree.forEach(rt => {
			if (rt.propInternalValue == collabSpace.dbValue) {
				if (orgs.find(org => org.propInternalValue === rt.org.propInternalValue) == undefined) {
					orgs.push(rt.org);
				}
			}

		});
	}
	return orgs;
};

export let getAllRolesForOrgNCollabSpace = function (collabSpace, org) {
	let roles = [];

	if (collabSpace == undefined || org == undefined) {
		return roles;
	}

	if (expData && expData.roleTree) {
		expData.roleTree.forEach((rt) => {
			if (rt.propInternalValue == collabSpace.dbValue) {
				if (rt.org && (rt.org.propInternalValue === org.dbValue)) {
					if (roles.length < 1 || (roles.find(rl => rl.propInternalValue == rt.role.propInternalValue) == undefined)) {
						roles.push(rt.role);
					}
				}
			}

		});
	}
	return roles;
};



/**
 * Get the Data from 3DEXPERIENCE and load in the Dialog
 * @param {Object} data - panel data
 */
export let initializeRoleIn3DXData = async function (input) {
	// let collaborativeSpaceValues = [];
	// let organizationValues = [];
	// let roleValues = [];
	let failedToRetriveUserRoleInfoMsg = input.failedToRetriveUserRoleInfoMsg;
	// let currentuser = appCtxSvc.getCtx('user');
	// let user_name = currentuser.props.user_name.dbValue;

	if (expData == undefined || expData.expUser == undefined) {
		messagingSvc.showWarning('3DEXPERIENCE user is required to get this information, please provide a valid 3DEXPERIENCE User!');
		return;
	}

	if (expData.roleInfoUpdated) {
		return expData;
	}

	expData.roleInfoUpdated = false;
	expData.defaultRole = {};
	expData.selectedRole = {};
	expData.defaultOrg = {};
	expData.selectedOrg = {};
	expData.defaultCollabSpace = {};
	expData.selectedCollabSpace = {};

	expData.roleTree = [];
	expData.collaborativeSpaceValues = [];
	expData.organizationValues = [];
	expData.roleValues = [];

	let parameters = new Map();
	parameters.set(ATTRIBUTE_NAME_USER, expData.expUser.dbValue);
	let requestXmlDocument = createRequestXMLDocument(GET_ROLE_INFO_FOR_NAME, parameters);

	await makeRequestToTCI(requestXmlDocument, false).then((responseXml) => {
		// for debug
		// const serializer = new XMLSerializer();
		// const xmlStr = serializer.serializeToString(responseXml);
		// console.info(xmlStr);

		let attributeElements = responseXml.getElementsByTagName(ATTRIBUTE);
		if (attributeElements.length > 0) {
			let prefCredentialsStr, queryTypesStr, queryTypesDispStr;

			for (let x = 0; x < attributeElements.length; x++) {
				if (attributeElements[x].hasAttributes()) {

					switch (attributeElements[x].getAttribute(NAME)) {
						case ENOVIA_PREFERRED_CREDENTIALS:
							prefCredentialsStr = attributeElements[x].getAttribute(VALUE);
							break;

						case ENOVIA_QUERY_TYPE_NAME_LIST:
							queryTypesStr = attributeElements[x].getAttribute(VALUE);
							break;

						case ENOVIA_QUERY_TYPE_DISPLAY_LIST:
							queryTypesDispStr = attributeElements[x].getAttribute(VALUE);
							break;
					}
				}
			}

			if (prefCredentialsStr) {
				const creds = prefCredentialsStr.split(".");
				if (creds.length == 3) {
					expData.roleInfoUpdated = true;

					expData.defaultRole.dbValue = creds[0];
					expData.selectedRole.dbValue = creds[0];

					expData.defaultOrg.dbValue = creds[1];
					expData.selectedOrg.dbValue = creds[1];

					expData.defaultCollabSpace.dbValue = creds[2];
					expData.selectedCollabSpace.dbValue = creds[2];
				}

			}
			if (queryTypesStr && queryTypesDispStr) {
				const queryTypeDBNames = queryTypesStr.split(LIST_SEPARATOR);
				const queryTypeDispNames = queryTypesDispStr.split(LIST_SEPARATOR);

				if (queryTypeDBNames.length > 0 && queryTypeDBNames.length == queryTypeDispNames.length) {
					let roleTree = [];
					for (let i = 0; i < queryTypeDBNames.length; ++i) {
						// queryTypeMap.set(queryTypeDBNames[i], queryTypeDispNames[i]);
						const dbCreds = queryTypeDBNames[i].split(".");
						const dispCreds = queryTypeDispNames[i].split(".");
						if (dbCreds.length == 3 && dispCreds.length == 3) {
							let dbRole = dbCreds[0];
							let dbOrg = dbCreds[1];
							let dbCollabSpace = dbCreds[2];

							let dispRole = dispCreds[0];
							let dispOrg = dispCreds[1];
							let dispCollabSpace = dispCreds[2];

							let cs = {
								propInternalValue: dbCollabSpace,
								propDisplayValue: dispCollabSpace,
								org: {
									propInternalValue: dbOrg,
									propDisplayValue: dispOrg
								},
								role: {
									propInternalValue: dbRole,
									propDisplayValue: dispRole
								}
							};

							roleTree.push(cs);

							if (expData.defaultRole.dbValue === dbRole
								&& expData.defaultOrg.dbValue === dbOrg
								&& expData.defaultCollabSpace.dbValue === dbCollabSpace) {

								expData.defaultRole.uiValue = dispRole;
								expData.selectedRole.uiValue = dispRole;

								expData.defaultOrg.uiValue = dispOrg;
								expData.selectedOrg.uiValue = dispOrg;

								expData.defaultCollabSpace.uiValue = dispCollabSpace;
								expData.selectedCollabSpace.uiValue = dispCollabSpace;
							}
						}
					}
					expData.roleTree = roleTree;

					if (expData.roleTree) {
						expData.roleTree.forEach(rt => {
							if (expData.collaborativeSpaceValues.length < 1 ||
								(expData.collaborativeSpaceValues.find(cs => cs.propInternalValue === rt.propInternalValue) == undefined)) {
								expData.collaborativeSpaceValues.push(rt);
							}
						});


						if (expData.defaultCollabSpace && expData.defaultCollabSpace != "") {
							expData.organizationValues = getAllOrgForCollabSpace(expData.defaultCollabSpace);
							if (expData.defaultOrg && expData.defaultOrg != "") {
								expData.roleValues = getAllRolesForOrgNCollabSpace(expData.defaultCollabSpace, expData.defaultOrg);
							}
						}

					}
				}
			}

		} else {
			// closeCurrentPanel();
			messagingSvc.showWarning("Response XML do not have any Attribute elements.");
		}

	}).catch((errMsg) => {
		// closeCurrentPanel();
		popupService.close();

		if (errMsg.includes('[1140]')) {
			messagingSvc.showError(failedToRetriveUserRoleInfoMsg + expData.expUser.dbValue);
		} else {
			messagingSvc.showError(errMsg);
		}
	});

	return expData;
};

export let storeExpUser = async function (username) {
	if (expData && expData.expUser) {
		expData.expUser.dbValue = username;
		expData.expUser.dispValue = username;
	}
};

export let storeExpSecret = async function (secret) {
	if (expData && expData.expPassword) {
		expData.expPassword.dbValue = secret;
		expData.expPassword.dispValue = secret;
	}
};


export let use3DXSelectedRole = (inputRole) => {
	return inputRole;
};

export let useSelectedRoleInLoginWindow = (collabSpace, org, role) => {
	if (collabSpace && org && role) {
		let input = 'ctx::' + role + '.' + org + '.' + collabSpace;
		eventBus.publish('loginTo3DExp.updateRoleFrom3DX', input);
	}
	popupService.close();
};

export let collaborationSpaceChanged = function (collabSpace, org, role) {
	// if (expData.selectedCollabSpace && expData.selectedCollabSpace.dbValue === collabSpace.dbValue && expData.selectedOrg && expData.selectedOrg.dbValue === org.dbValue && expData.selectedRole && expData.selectedRole.dbValue === role.dbValue) {
	if (expData.selectedCollabSpace?.dbValue === collabSpace.dbValue) {
		return expData;
	}

	expData.selectedCollabSpace.dbValue = collabSpace.dbValue;
	expData.selectedCollabSpace.uiValue = collabSpace.uiValue;

	expData.organizationValues = getAllOrgForCollabSpace(collabSpace);
	expData.roleValues = getAllRolesForOrgNCollabSpace(collabSpace, org);

	if (expData.organizationValues.find(o => o.propInternalValue === org.dbValue)) {
		organizationChanged(collabSpace, org, role);
	} else if (expData.organizationValues && expData.organizationValues.length > 0) {
		expData.selectedOrg.dbValue = expData.organizationValues[0].propInternalValue;
		expData.selectedOrg.uiValue = expData.organizationValues[0].propDisplayValue;

		organizationChanged(collabSpace, expData.selectedOrg, role);

	} else {
		clearProperty(expData.selectedOrg);
		clearProperty(expData.selectedRole);
	}

	return expData;
};

export let organizationChanged = function (collabSpace, org, role) {
	expData.selectedOrg.dbValue = org.dbValue;
	expData.selectedOrg.uiValue = org.uiValue;

	expData.roleValues = getAllRolesForOrgNCollabSpace(collabSpace, org);
	if (expData.roleValues.find(r => r.propInternalValue === role.dbValue)) {
		//Skip Reset as the current value is a valid one
	} else if (expData.roleValues.length > 0) {
		expData.selectedRole.dbValue = expData.roleValues[0].propInternalValue;
		expData.selectedRole.uiValue = expData.roleValues[0].propDisplayValue;
	} else {
		clearProperty(expData.selectedRole);
	}

	return expData;
};

export let roleChanged = function (collabSpace, org, role) {
	expData.selectedRole.dbValue = role.dbValue;
	expData.selectedRole.uiValue = role.uiValue;
};

/**
  * Reload productsIn3DX Page
  */
export let reloadPageProductsIn3DX = function () {
	var action = {
		actionType: 'Navigate',
		navigateTo: 'productsIn3DX',
		navigationParams: {}
	};
	navigationSvc.navigate(action, action.navigationParams);
};

/**
  * Navigate to the input URL in new tab
  * @param {String} urlString URL string that need to open
  * @param {Object} navContext URL JSON Object to pass as Navigation Context
  */
export let navigateToURL = function (urlString) {
	if (urlString) {
		var action = {
			actionType: 'Navigate',
			navigateTo: urlString,
			navigationParams: {},
			navigateIn: 'newTab',
		};
		navigationSvc.navigate(action, action.navigationParams);
	}
};
// export let navigateToURL = function (urlString, navContext) {
// 	if (urlString) {
// 		var action = {
// 			actionType: 'Navigate',
// 			navigateTo: urlString,
// 			navigationParams: {
// 				navigationContext: navContext
// 			},
// 			navigateIn: 'newTab',
// 		};
// 		navigationSvc.navigate(action, action.navigationParams);
// 	}
// };

/**
 * Clear the cache data.
 */
const clearCachedData = (itemName) => {
	localStrg.removeItem(itemName);
};

/**
 * Update this model object's awp properties, used only to create 3DEXPERIENCE object
 * awp0CellProperties.
 * @param {ViewModelObject} viewModelObject - The 3DEXPERIENCE VMO to update properties on.
 * @param {Object} sourceObject - The Source Object to copy values from
 */
var updatePropertiesOn3DXVMO = function (viewModelObject, sourceObject) {
	if (sourceObject && viewModelObject && sourceObject.props) {
		let propNames = Object.keys(sourceObject.props);

		propNames.forEach((pn) => {
			// create missing properties which are not created because the value was empty
			if (!viewModelObject.props[pn]) {
				let vmProp = viewModelObjectService.constructViewModelProperty('', pn, viewModelObject, false);
				viewModelObject.props[pn] = vmProp;
			}

			//Set Property Attributes
			uwPropertySvc.setIsArray(viewModelObject.props[pn], false);

			// Set the Display Name/Label of Property
			uwPropertySvc.setPropertyDisplayName(viewModelObject.props[pn], pn);
			uwPropertySvc.setPropertyLabelDisplay(viewModelObject.props[pn], 'PROPERTY_LABEL_AT_SIDE');

			// Set the created and modified date properties type as Date and the value
			if (pn == 'created' || pn == 'modified') {
				uwPropertySvc.setDataType(viewModelObject.props[pn], 'DATE');
				let dateObj = new Date(Date.parse(sourceObject.props[pn]));
				let dateDispValue = dateTimeSvc.formatSessionDateTime(dateObj);
				uwPropertySvc.setValue(viewModelObject.props[pn], dateObj);
				uwPropertySvc.setDisplayValue(viewModelObject.props[pn], [dateDispValue]);

			} else {
				//Set other properties type as string and the value
				uwPropertySvc.setDataType(viewModelObject.props[pn], 'STRING');
				uwPropertySvc.setValue(viewModelObject.props[pn], (sourceObject.props[pn] || ''));
				uwPropertySvc.setDisplayValue(viewModelObject.props[pn], ([sourceObject.props[pn]] || ''));
			}


			//Set Properties disabled so they will not be displayed in UI
			if (pn == 'PLM_ExternalID') {
				uwPropertySvc.setIsEnabled(viewModelObject.props[pn], false);
				uwPropertySvc.setIsDisplayable(viewModelObject.props[pn], false);
			}

		});
	}
}


/**
 * Update this model object's awp cell properties, used only to create 3DEXPERIENCE object
 * awp0CellProperties.
 *
 * @param {ViewModelObject} viewModelObject - The 3DEXPERIENCE VMO to update properties on.
 */
var updateCellPropertiesOn3DXVMO = function (viewModelObject) {
	if (viewModelObject.props) {
		viewModelObject.cellProperties = {};
		viewModelObject.cellHeader1 = viewModelObject.props.Title.dbValue || '';
		viewModelObject.cellHeader2 = (viewModelObject.type || '')
			+ ' | ' + (viewModelObject.props.Revision.dbValue || '')
			+ ' | ' + (viewModelObject.props.State.dbValue || '')
			+ ' | ' + (viewModelObject.props.Owner.dbValue || '');

		viewModelObject.cellProperties['Name'] = {
			key: 'Name',
			value: (viewModelObject.props.Name.dbValue || '')
		};

		// viewModelObject.cellProperties['Owner'] = {
		// 	key: 'Owner',
		// 	value: (viewModelObject.props.Owner.dbValue || '')
		// };
	}
};

/**
 * Update 3DEXPERIENCE Object Icon URL
 *
 * @param {ViewModelObject} viewModelObject - The 3DEXPERIENCE VMO to update Icon on.
 */
var updateIconOn3DXVMO = function (viewModelObject) {
	if (viewModelObject) {
		viewModelObject.thumbnailURL = '';
		viewModelObject.hasThumbnail = false;

		switch (viewModelObject.type) {
			// case 'iPLMPart':
			// 	viewModelObject.typeIconURL = 'assets/image/typeItemRevision48.svg';
			// 	break;

			// case 'VPMReference':
			// 	viewModelObject.typeIconURL = 'assets/image/typePartRevision48.svg';
			// 	break;

			default:
				viewModelObject.typeIconURL = 'assets/image/typeVpmReference48.svg';

		}
	}
};

/**
 * Filtering 3DEXPERIENCE Objects based on provided property with the 'Contains' operation.
 *
 * @param {Array} objectDatas The dataset to filter
 * @param {string} filterText Filter to apply
 * @param {string} filterProperty Filter to apply on property, default Title
 * @returns {Object} The objData data that matches the filter
 */
let processContainsFilter = function (objectDatas, filterText, filterProperty = 'Title') {
	if (filterText && filterText.trim().length > 0 && filterText.trim() != '*') {
		return objectDatas.filter(function (objData) {
			if (objData.props[filterProperty].uiValue) {
				return objData.props[filterProperty].uiValue.toLowerCase().includes(filterText.toLowerCase());
			} else if (objData.props[filterProperty].uiValues) {
				return objData.props[filterProperty].uiValues.toString().toLowerCase().includes(filterText.toLowerCase());
			}
		});
	}

	return objectDatas;
};

export let loadQueriedProductsFrom3DX = (input) => {
	//TODO: Pending implementation
	let startIndex = input.searchInput.startIndex || 0;
	let pageSize = input.searchInput.maxToLoad || 50;
	let filterText = input.searchInput.searchCriteria.searchString;
	let dataProviderName = input.dataProvider.name;
	let filterProperty = 'Title';

	let endIndex = startIndex + pageSize;
	let objectDatas = processContainsFilter(input.searchResults, filterText, filterProperty);
	input.searchResults = objectDatas.slice(startIndex, endIndex);

	input.dataProvider.selectionModel.setMode('single');
	let newSearchData = { ...input.subPanelContext.searchState.value };
	newSearchData.searchResults = objectDatas.slice(startIndex, endIndex);
	newSearchData.totalFound = input.totalFound;
	newSearchData.totalLoaded = newSearchData.searchResults.length;
	newSearchData.dataProvider = input.dataProvider;

	input.subPanelContext.searchState.update(newSearchData);
	return newSearchData;


	// if(input.searchResults && input.searchResults.length > 0){
	// 	let eventData = {
	// 		selectedObjects: [input.searchResults[0]], 
	// 		selectedUids: [input.searchResults[0].uid]
	// 	};
	// 	eventBus.publish( dataProviderName + '.selectionChangeEvent', eventData );
	// }

	// return input;
};

export let initializeQueriedProductsFrom3DX_states = () => {
	var deferred = AwPromiseService.instance.defer();
	// Currently this is just a dummy function called from states.json
	// This must not bew removed.
	// if removed the initializeQueriedProductsFrom3DX function will not get called on page refresh
	deferred.resolve();
};

export let initializeQueriedProductsFrom3DX = async () => {
	var result = {};
	result.data = [];
	let envObjects = [];

	// let isTciLoggedIn = appCtxSvc.getCtx(IS_TCI_LOGGED_IN);
	if (!expData) {
		updateExpDataFromLocalStorage();
	}

	if (!expQueryInput) {
		const queryDataJson = localStrg.get('tciData.queryInput');
		expQueryInput = JSON.parse(queryDataJson);
		// clearCachedData('tciData.queryInput');
	}

	envObjects = await performQueryIn3DX(expQueryInput);

	// if (isTciLoggedIn === true && activeLoc === 'tciLocation') {
	// 	let queryInput = appCtxSvc.getCtx('tciData.queryInput');
	// 	envObjects = performQueryIn3DX(queryInput);
	// }
	// else {

	// 	// appCtxSvc.registerCtx(IS_TCI_LOGGED_IN, true);
	// 	// setTciLoggedIn();

	// 	// const expDataJson = localStrg.get('expData');
	// 	// expData = JSON.parse(expDataJson);
	// 	// clearCachedData('expData');

	// 	const queryInputJson = localStrg.get('queryInput');
	// 	let queryInput = JSON.parse(queryInputJson);
	// 	appCtxSvc.registerCtx('tciData.queryInput', queryInput);
	// 	clearCachedData('queryInput');

	// 	const queryDataFrom3DXJson = localStrg.get('queryDataFrom3DX');
	// 	if (queryDataFrom3DXJson) {
	// 		envObjects = JSON.parse(queryDataFrom3DXJson);
	// 		clearCachedData('queryDataFrom3DX');
	// 	}
	// }

	if (envObjects && envObjects.length > 0) {
		for (var x = 0; x < envObjects.length; x++) {
			const vmo = viewModelObjectService.constructViewModelObject(envObjects[x]);
			updatePropertiesOn3DXVMO(vmo, envObjects[x]);
			updateCellPropertiesOn3DXVMO(vmo);
			updateIconOn3DXVMO(vmo);
			result.data.push(vmo);
		}
	}
	else {
		messagingSvc.showWarning('No Results found for provided search criteria!');
	}

	appCtxSvc.updateCtx('tciData.queryDataFrom3DX', result.data);
};

/**
 * Synchronously execute the 'setSelection' API on the given {UwSelectionModel} object.
 *
 * @param {UwSelectionModel} selectionModel - The {UwSelectionModel} object to set the selection on.
 * @param {String|String[]|Object|Object[]} newSelection - What to set as selection
 */
export let setSelectionOnSelectionModel = function (selectionModel, newSelection) {
	selectionModel.setSelection(newSelection);
};

export let getSelectedObjForQueriedProductsFrom3DX = (input) => {
	let subPanelContext = input.subPanelContext;

	if (subPanelContext && subPanelContext.selection && subPanelContext.selection.length > 0) {
		return subPanelContext.selection[0];
	}

	return {};
}

export let checkUseOfIncludeEffectivity = async (selectedObject) => {
	let type = selectedObject.props.Type.dbValue;
	let name = selectedObject.props.Name.dbValue;
	let majorrevision = selectedObject.props.majorrevision.dbValue;
	let settingName = 'Show3DXSendConfigurationEnabled';
	let result = false;

	let parameters = getCredentialsMap(expData.loggedUser.user.dbValue, expData.loggedUser.secret.dbValue, expData.loggedUser.role.dbValue, expData.loggedUser.vault.dbValue, expData.loggedUser.isPasswordEncrypted);
	parameters.set('type', type);
	parameters.set('name', name);
	parameters.set('revision', majorrevision);

	let requestXmlDocument = createRequestXMLDocument(GET_SETTINGS_INFO, parameters);

	await makeRequestToTCI(requestXmlDocument, false).then((responseXml) => {

		//for debug
		// const serializer = new XMLSerializer();
		// const xmlStr = serializer.serializeToString(responseXml);
		// console.info(xmlStr);

		let attrElements = responseXml.getElementsByTagName(settingName);
		if (attrElements.length > 0) {
			//Collecte all attribute element names and values
			for (let x = 0; x < attrElements.length; x++) {
				let value = attrElements[x].getAttribute('value');
				if (value.toLowerCase() == 'true') {
					result = true;
					break;
				}
			}
		}

	}).catch((errMsg) => {
		messagingSvc.showError(errMsg);
	});

	return result;
};


export let retrieveProductFrom3DExp = (input) => {
	let type = input.selectedObject.props.Type.dbValue;
	let name = input.selectedObject.props.Name.dbValue;
	let majorrevision = input.selectedObject.props.majorrevision.dbValue;

	let parameters = getCredentialsMap(expData.loggedUser.user.dbValue, expData.loggedUser.secret.dbValue, expData.loggedUser.role.dbValue, expData.loggedUser.vault.dbValue, expData.loggedUser.isPasswordEncrypted);
	parameters.set('type', type);
	parameters.set('name', name);
	parameters.set('revision', majorrevision);

	// default is all
	let expandLevel = -1;
	if (input.data.expDepthProp.dbValue == 'all') {
		expandLevel = -1;
	}
	else if (input.data.expDepthProp.dbValue == 'none') {
		expandLevel = 0;
	}
	else if (input.data.expDepthProp.dbValue == 'selectDepth') {
		expandLevel = input.data.depthLevel.dbValue;
	}

	parameters.set("expand_level", expandLevel.toString());
	parameters.set("force", input.data.optionForce.dbValue.toString());
	parameters.set("remove_existing", input.data.optionRemoveNonExistingChildren.dbValue.toString());
	parameters.set("TRANSFER_SUBSTRUCTURES_PARALLEL", input.data.transferInParallel.dbValue.toString());
	parameters.set("send_configuration", input.data.optionIncludeEffectivity.dbValue.toString());

	let requestXmlDocument = createRequestXMLDocument(SEND_ENOVIA_OBJECT_TO_TC, parameters);

	// just for debug
	// const serializer = new XMLSerializer();
	// const xmlStr = serializer.serializeToString(requestXmlDocument);
	// console.info(xmlStr);

	makeRequestToTCI(requestXmlDocument, true, true);
};

export let updateSelectionModel = (input) => {
	input.dataProvider.selectionModel.setMode('single');
	let newSearchData = { ...input.subPanelContext.searchState.value };
	newSearchData.dataProvider = input.dataProvider;

	input.subPanelContext.searchState.update(newSearchData);
	return newSearchData;
};

export let testing = (input, input2, input3, input4) => {
	console.log("In testing function...");
	return input;
}


/**
* This function sends the request.
*/
export let openIn3DX = function (ctx) {
	adapterSvc.getAdaptedObjects(ctx.mselected).then(function (adaptedObjs) {
		if (adaptedObjs.length > 0) {
			let selectedObj = adaptedObjs[0];

			let parameters = new Map();
			parameters.set("UID", selectedObj.uid);
			parameters.set("getPhysicalid", true);

			let requestXmlDocument = createRequestXMLDocument("OpenIn3DX", parameters);

			makeRequestToTCI(requestXmlDocument, false, false).then(async (responseXml) => {
				if (responseXml) {
					//for debug
					const serializer = new XMLSerializer();
					const xmlStr = serializer.serializeToString(responseXml);
					console.log(xmlStr);

					//TODO: read TC preferences
					let baseUrl = await getPreferenceValueOrDefault('XCI_3DX_DASHBOARD_URL');// 'https://nis07s209.epdm-d.edm.dsh.de:21544/3ddashboard'
					let appId = await getPreferenceValueOrDefault('XCI_3DX_DASHBOARD_APP', 'ENOSCEN_AP');
					let envId = await getPreferenceValueOrDefault('XCI_3DX_TENANT', 'OnPremise');

					if (!baseUrl) {
						messagingSvc.showError('Please verify if the Teamcenter Preference XCI_3DX_DASHBOARD_URL is set or contact Teamcenter Administrator.');
						return;
					}

					let attrElements = responseXml.getElementsByTagName(ATTRIBUTE);
					if (attrElements.length > 0) {
						let objectId = '';
						let objectType = '';
						for (let x = 0; x < attrElements.length; x++) {
							let attrName = attrElements[x].getAttribute('name').toLowerCase();
							if (attrName == 'type') {
								objectType = attrElements[x].getAttribute('value');
							}
							else if (attrName == 'physicalid') {
								objectId = attrElements[x].getAttribute('value');
							}
						}

						let uriComp = encodeURIComponent(`{"data":{"items":[{"objectId":"${objectId}","objectType":"${objectType}","envId":"${envId}"}]}}`);

						let url = `${baseUrl}/#app:${appId}/content:X3DContentId=${uriComp}`;

						navigateToURL(url);


					} else {
						messagingSvc.showInfo('No response received from 3DEXPERIENCE for object ' + selectedObj.props.object_string.dbValues[0]);
					}

				}

			}).catch((errMsg) => {
				messagingSvc.showError(errMsg);
			});

		}
	});

};


export default exports = {
	testing,
	getRules,
	getItemRevisions,
	getUid,
	getExpand,
	sendTo3DX,
	sendFolderTo3DX,
	showAboutTCI,
	initializeDataFromLocalStorage,
	updateSelectionAndActivateSendTo3DXPanel,
	alert: alert2,
	loginTo3DExp,
	updateTciLogginFromLocalStorage,
	populate3DExpLoginInfo,
	fetchPropInfoAndActivateUpdatePropsIn3DXPanel,
	initializeVMPDataForViewPropsIn3DX,
	initializeVMPDataForUpdatePropsIn3DX,
	performUpdatePropsIn3DExp,
	clearProperty,
	initializeRoleIn3DXData,
	storeExpUser,
	storeExpSecret,
	collaborationSpaceChanged,
	organizationChanged,
	roleChanged,
	useSelectedRoleInLoginWindow,
	use3DXSelectedRole,
	logoutOf3DX,
	fetchPropsAndActivateViewPropsFrom3DExpPanel,
	fetchProcessDataFromTci,
	loadTciProcessData,
	loadTciProcessColumns,
	activateQueryProductsIn3DXPanel,
	queryIn3DExpAndActivateProductsIn3DXPage,
	navigateToURL,
	updateFieldWithDefaultValues,
	initializeQueriedProductsFrom3DX_states,
	initializeQueriedProductsFrom3DX,
	loadQueriedProductsFrom3DX,
	getSelectedObjForQueriedProductsFrom3DX,
	setSelectionOnSelectionModel,
	retrieveProductFrom3DExp,
	checkUseOfIncludeEffectivity,
	updateSelectionModel,
	openIn3DX

};

app.factory('tciService', () => exports);

